"""Example server code.

Make a fruit API.
"""

import asyncio
import logging
from typing import Any, Dict

from rest_tools.server import (  # noqa: E402 # pylint: disable=C0413,E0401
    RestHandler,
    RestHandlerSetup,
    RestServer,
    role_authorization,
)
from rest_tools.utils.json_util import (  # noqa: E402 # pylint: disable=C0413,E0401
    json_decode,
)
from wipac_telemetry import tracing_tools  # type: ignore[import]


class FruitsHanlder(RestHandler):
    """Handle Fruit requests."""

    def initialize(self, fruit: Dict[str, Any], *args: Any, **kwargs: Any) -> None:  # type: ignore
        """Initialize."""
        super().initialize(*args, **kwargs)
        self.fruit = fruit  # pylint: disable=W0201

    @role_authorization(roles=["admin", "user"])
    async def get(self) -> None:
        """Write existing fruits."""
        assert tracing_tools.get_current_span().parent.span_id  # type: ignore[attr-defined]

        logging.info("fruits: %r", self.fruit)

        self.write({"fruits": self.fruit})

    @role_authorization(roles=["admin"])
    async def post(self) -> None:
        """Handle a new fruit."""
        assert tracing_tools.get_current_span().parent.span_id  # type: ignore[attr-defined]

        body = json_decode(self.request.body)
        logging.info("body: %r", body)

        self.fruit[body["name"]] = body

        self.write({})


def main() -> None:
    """Establish and run REST server."""
    logging.basicConfig(level=logging.DEBUG)

    args = RestHandlerSetup({"auth": {"secret": "secret"}, "debug": True})
    args["fruit"] = {}  # this could be a DB, but a dict works for now

    server = RestServer(debug=True)
    server.add_route("/api/fruits", FruitsHanlder, args)
    server.startup(address="localhost", port=8080)

    asyncio.get_event_loop().run_forever()


if __name__ == "__main__":
    main()



"""Test client.utils.request_and_validate module."""

import re
from typing import AsyncIterator, Callable

import openapi_core
import pytest
import pytest_asyncio
import requests
import tornado
from jsonschema_path import SchemaPath
from openapi_core.templating.paths.exceptions import PathNotFound
from openapi_core.templating.responses.exceptions import ResponseNotFound
from openapi_core.validation.response.exceptions import DataValidationError

from rest_tools.client import RestClient
from rest_tools.client.utils import request_and_validate
from rest_tools.server import RestServer, RestHandler


@pytest_asyncio.fixture
async def server(port: int) -> AsyncIterator[Callable[[], RestClient]]:
    """Start up REST server and attach handlers."""
    rs = RestServer(debug=True)

    class TestHandler(RestHandler):
        ROUTE = "/echo/this"

        async def post(self) -> None:
            if self.get_argument("raise", None):
                raise tornado.web.HTTPError(self.get_argument("raise"), "it's an error")
            self.write({"echo-echo": self.get_argument("echo", {})})

    rs.add_route(TestHandler.ROUTE, TestHandler)
    rs.startup(address="localhost", port=port)

    def client() -> RestClient:
        return RestClient(f"http://localhost:{port}", retries=0)

    try:
        yield client
    finally:
        await rs.stop()


OPENAPI_SPEC = openapi_core.OpenAPI(
    SchemaPath.from_dict(
        {
            "openapi": "3.1.0",
            "info": {
                "title": "Test Schema",
                "summary": "A schema for testing",
                "description": "This is a test description",
                "contact": {
                    "name": "WIPAC Developers",
                    "url": "icecube.wisc.edu",
                    "email": "developers@icecube.wisc.edu",
                },
                "license": {"name": "MIT License"},
                "version": "0.0.0",
            },
            "components": {
                "parameters": {},
                "schemas": {
                    "EchoObject": {
                        "type": "object",
                        "properties": {
                            "foo": {"type": "integer"},
                        },
                        "required": ["foo"],
                        "additionalProperties": False,
                    },
                },
            },
            "paths": {
                "/echo/this": {
                    # "parameters": [],
                    "post": {
                        # "requestBody": {},  # NOTE: these tests don't include input validation, so don't bother
                        "responses": {
                            "200": {
                                "description": "echoing this value",
                                "content": {
                                    "application/json": {
                                        "schema": {
                                            "type": "object",
                                            "properties": {
                                                "echo-echo": {
                                                    "$ref": "#/components/schemas/EchoObject"
                                                },
                                            },
                                        }
                                    }
                                },
                            },
                            "400": {
                                "description": "invalid request arguments",
                                "content": {
                                    "application/json": {
                                        "schema": {
                                            "type": "object",
                                            "properties": {
                                                "code": {
                                                    "description": "http error code",
                                                    "type": "integer",
                                                },
                                                "error": {
                                                    "description": "http error reason",
                                                    "type": "string",
                                                },
                                            },
                                            "required": ["code", "error"],
                                            "additionalProperties": False,
                                        }
                                    }
                                },
                            },
                        },
                    },
                },
            },
        }
    )
)


async def test_000__valid(server: Callable[[], RestClient]) -> None:
    """Test valid request data."""
    rc = server()

    # validate response data
    res = await request_and_validate(
        rc, OPENAPI_SPEC, "POST", "/echo/this", {"echo": {"foo": 123}}
    )
    assert res == {"echo-echo": {"foo": 123}}

    # validate response error
    with pytest.raises(requests.HTTPError) as e:
        await request_and_validate(
            rc,
            OPENAPI_SPEC,
            "POST",
            "/echo/this",
            {"raise": 400},
        )
    assert e.value.response.status_code == 400


async def test_010__invalid(server: Callable[[], RestClient]) -> None:
    """Test invalid request data."""
    rc = server()

    with pytest.raises(
        PathNotFound, match=re.escape(f"Path not found for {rc.address}/foo/bar/baz")
    ):
        await request_and_validate(rc, OPENAPI_SPEC, "GET", "/foo/bar/baz")

    # validate response data

    with pytest.raises(
        DataValidationError,
        match=re.escape("Failed to cast value to object type: 123"),
    ):
        await request_and_validate(
            rc, OPENAPI_SPEC, "POST", "/echo/this", {"echo": 123}
        )

    with pytest.raises(
        DataValidationError,
        match=re.escape("Failed to cast value to integer type: hello"),
    ):
        await request_and_validate(
            rc, OPENAPI_SPEC, "POST", "/echo/this", {"echo": {"foo": "hello"}}
        )

    with pytest.raises(
        DataValidationError,
        match=re.escape(
            "Value {'echo-echo': {}} not valid for schema of type object: (<ValidationError: \"'foo' is a required property\">,)"
        ),
    ):
        await request_and_validate(rc, OPENAPI_SPEC, "POST", "/echo/this", {"baz": 123})

    # validate response error
    with pytest.raises(
        ResponseNotFound, match=re.escape("Unknown response http status: 401")
    ):
        await request_and_validate(
            rc,
            OPENAPI_SPEC,
            "POST",
            "/echo/this",
            {"raise": 401},
        )



"""Test script for RestClient."""

# fmt:quotes-ok

import json
import logging
import re
import signal
from contextlib import contextmanager
from typing import Any, Iterable, Iterator
from unittest.mock import Mock

import pytest
import urllib3
from httpretty import HTTPretty, httprettified  # type: ignore[import]
from requests import PreparedRequest
from requests.exceptions import SSLError, Timeout
from rest_tools.client import (
    MAX_RETRIES,
    CalcRetryFromBackoffMax,
    CalcRetryFromWaittimeMax,
    RestClient,
)
from rest_tools.utils.json_util import json_decode, json_encode

logger = logging.getLogger("rest_client")


def test_001_init() -> None:
    """Test `__init__()` & `close()`."""
    rpc = RestClient("http://test", "passkey")
    rpc.close()


@pytest.mark.asyncio
async def test_010_request(requests_mock: Mock) -> None:
    """Test `async request()`."""
    result = {"result": "the result"}
    rpc = RestClient("http://test", "passkey", timeout=0.1)

    def response(req: PreparedRequest, ctx: object) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        _ = json_decode(req.body)
        return json_encode(result).encode("utf-8")

    requests_mock.post("/test", content=response)
    ret = await rpc.request("POST", "test", {})

    assert requests_mock.called
    auth_parts = requests_mock.last_request.headers['Authorization'].split(' ', 1)
    assert auth_parts[0].lower() == 'bearer'
    assert auth_parts[1] == 'passkey'
    assert ret == result

    result2 = {"result2": "the result 2"}

    def response2(req: PreparedRequest, ctx: object) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        _ = json_decode(req.body)
        return json_encode(result2).encode("utf-8")

    requests_mock.post("/test2", content=response2)
    ret = await rpc.request("POST", "/test2")

    assert requests_mock.called
    assert ret == result2


@pytest.mark.asyncio
async def test_011_request(requests_mock: Mock) -> None:
    """Test request in `async request()`."""
    rpc = RestClient("http://test", "passkey", timeout=0.1)
    requests_mock.get("/test", content=b"")
    ret = await rpc.request("GET", "test", {})

    assert requests_mock.called
    assert ret is None


@pytest.mark.asyncio
async def test_020_timeout(requests_mock: Mock) -> None:
    """Test timeout in `async request()`."""
    rpc = RestClient("http://test", "passkey", timeout=0.1)
    requests_mock.post("/test", exc=Timeout)

    with pytest.raises(Timeout):
        _ = await rpc.request("POST", "test", {})


@pytest.mark.asyncio
async def test_021_ssl_error(requests_mock: Mock) -> None:
    """Test ssl error in `async request()`."""
    rpc = RestClient("http://test", "passkey", timeout=0.1)
    requests_mock.post("/test", exc=SSLError)

    with pytest.raises(SSLError):
        _ = await rpc.request("POST", "test", {})


@pytest.mark.asyncio
async def test_022_request(requests_mock: Mock) -> None:
    """Test `async request()`."""
    rpc = RestClient("http://test", "passkey", timeout=0.1)
    requests_mock.get("/test", content=b'{"foo"}')

    with pytest.raises(Exception):
        _ = await rpc.request("GET", "test", {})


@pytest.mark.asyncio
async def test_030_request(requests_mock: Mock) -> None:
    """Test `async request()` with headers."""
    rpc = RestClient("http://test", "passkey", timeout=0.1)
    requests_mock.get("/test", content=b"")
    ret = await rpc.request("GET", "test", {}, {'foo': 'bar'})

    assert requests_mock.called
    assert requests_mock.last_request.headers['foo'] == 'bar'
    assert ret is None


@pytest.mark.asyncio
async def test_040_request_autocalc_retries() -> None:
    """Test auto-calculated retries options in `RestClient`."""
    for timeout, backoff_factor, arg, out in [
        (0.5, 0.75, 1, 1),
        (0.5, 0.75, MAX_RETRIES, MAX_RETRIES),
        #
        (0.5, 0.75, CalcRetryFromBackoffMax(0.1), 0),  # -1.90689 -> 0
        (0.5, 0.75, CalcRetryFromBackoffMax(0.5), 0),  # 0.41503
        (0.5, 0.75, CalcRetryFromBackoffMax(0.75), 1),  # 1.0
        (0.5, 0.75, CalcRetryFromBackoffMax(0.9), 1),  # 1.26303
        (0.5, 0.75, CalcRetryFromBackoffMax(8.1), 4),  # 4.43296
        #
        (0.5, 0.75, CalcRetryFromWaittimeMax(100), 7),  # 7.01484
        (15, 0.5, CalcRetryFromWaittimeMax(5 * 60), 8),  # 8.20825
        (1, 0.1, CalcRetryFromWaittimeMax(5 * 60), 11),  # 11.4854
        (5, 2, CalcRetryFromWaittimeMax(5 * 60), 7),  # 7.01635
    ]:
        print(f"{timeout=}, {backoff_factor=}, {arg=}, {out=}")
        rc = RestClient(
            "http://test",
            "passkey",
            timeout=timeout,
            retries=arg,  # type: ignore[arg-type]
            backoff_factor=backoff_factor,
        )
        assert rc.retries == out


@pytest.mark.asyncio
async def test_041_request_autocalc_retries_error() -> None:
    """Test auto-calculated retries options in `RestClient`."""
    for timeout, backoff_factor, arg in [
        (0.5, 0.75, MAX_RETRIES + 1),
        #
        (
            0.5,
            01e-07,
            CalcRetryFromBackoffMax(urllib3.util.retry.Retry.DEFAULT_BACKOFF_MAX),
        ),
        #
        (0.5, 0.3, CalcRetryFromWaittimeMax(60 * 60)),
    ]:
        print(f"{timeout=}, {backoff_factor=}, {arg=}")
        with pytest.raises(
            ValueError, match=re.escape(f"Cannot set # of retries above {MAX_RETRIES}")
        ):
            RestClient(
                "http://test",
                "passkey",
                timeout=timeout,
                retries=arg,  # type: ignore[arg-type]
                backoff_factor=backoff_factor,
            )


@pytest.mark.asyncio
async def test_042_request_autocalc_retries_error() -> None:
    """Test auto-calculated retries options in `RestClient`."""
    with pytest.raises(
        ValueError,
        match=r"CalcRetryFromBackoffMax\.backoff_max \(\d+\) cannot be greater than: urllib3\.util\.retry\.Retry\.DEFAULT_BACKOFF_MAX=\d+",
    ):
        RestClient(
            "http://test",
            "passkey",
            timeout=0.5,
            retries=CalcRetryFromBackoffMax(
                urllib3.util.retry.Retry.DEFAULT_BACKOFF_MAX + 1
            ),
            backoff_factor=0.001,
        )


def test_100_request_seq(requests_mock: Mock) -> None:
    """Test `request_seq()`."""
    result = {"result": "the result"}
    rpc = RestClient("http://test", "passkey", timeout=0.1)

    def response(req: PreparedRequest, ctx: object) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        _ = json_decode(req.body)
        return json_encode(result).encode("utf-8")

    requests_mock.post("/test", content=response)
    ret = rpc.request_seq("POST", "test", {})

    assert requests_mock.called
    assert ret == result


@pytest.mark.asyncio
async def test_101_request_seq(requests_mock: Mock) -> None:
    """Test `request_seq()`."""
    result = {"result": "the result"}
    rpc = RestClient("http://test", "passkey", timeout=0.1)

    def response(req: PreparedRequest, ctx: object) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        _ = json_decode(req.body)
        return json_encode(result).encode("utf-8")

    requests_mock.post("/test", content=response)
    ret = rpc.request_seq("POST", "test", {})

    assert requests_mock.called
    assert ret == result


@pytest.mark.asyncio
async def test_102_request_seq(requests_mock: Mock) -> None:
    """Test `request_seq()`."""
    rpc = RestClient("http://test", "passkey", timeout=0.1)

    def response(req: PreparedRequest, ctx: object) -> bytes:  # pylint: disable=W0613
        raise Exception()

    requests_mock.post("/test", content=response)

    with pytest.raises(Exception):
        rpc.request_seq("POST", "test", {})


@contextmanager
def _in_time(time, message):  # type: ignore[no-untyped-def]
    # Based on https://github.com/gabrielfalcao/HTTPretty/blob/master/tests/functional/test_requests.py#L290."""
    # A context manager that uses signals to force a time limit in tests
    # (unlike the `@within` decorator, which only complains afterward), or
    # raise an AssertionError.

    def handler(signum, frame):  # type: ignore[no-untyped-def]
        raise AssertionError(message)

    signal.signal(signal.SIGALRM, handler)
    signal.setitimer(signal.ITIMER_REAL, time)
    yield
    signal.setitimer(signal.ITIMER_REAL, 0)


def _jsonify(val: bytes) -> Any:
    return json.loads(val.strip()) if val.strip() else None


def _json_stream(iterable: Iterable[Any]) -> Iterator[Any]:
    for val in iterable:
        ret = _jsonify(val)
        if ret:  # no blanks
            yield ret


@httprettified  # type: ignore[misc]
def test_200_request_stream() -> None:
    """Test `request_stream()`.

    Based on https://github.com/gabrielfalcao/HTTPretty/blob/master/tests/functional/test_requests.py#L290
    """
    mock_url = "http://test"
    expected_stream = [
        b'{"foo-bar":"baz"}\n',
        b"\r\n",
        b"\n",
        b'{"green":["eggs", "and", "ham"]}\r\n',
        b'{"george": 1, "paul": 2, "ringo": 3, "john": 4}\r\n',
        b"\n",
        b'"this is just a string"\n',
        b"[1,2,3]",
    ]
    rpc = RestClient(mock_url, "passkey", timeout=1)

    json_stream = list(_json_stream(expected_stream))

    # test multiple times
    for test_num in range(2):
        print(f"\niteration #{test_num}")
        HTTPretty.register_uri(
            HTTPretty.POST,
            mock_url + "/stream/it/",
            body=(ln for ln in expected_stream),
            streaming=True,
        )
        response_stream = rpc.request_stream("POST", "/stream/it/", {})

        with _in_time(10, "Iterating by line is taking forever!"):
            for i, resp in enumerate(response_stream):
                print(f"resp={resp}")
                assert resp == json_stream[i]

    # now w/ chunk sizes
    for chunk_size in [None, -1, 0, 1, 2, 3, 4, 8, 9, 20, 100, 1024, 32768]:
        print(f"\nchunk_size: {chunk_size}")
        HTTPretty.register_uri(
            HTTPretty.POST,
            mock_url + "/stream/it/w/chunks",
            body=(ln for ln in expected_stream),
            streaming=True,
        )
        response_stream = rpc.request_stream(
            "POST", "/stream/it/w/chunks", {}, chunk_size=chunk_size
        )

        with _in_time(10, "Iterating by line is taking forever w/ chunks!"):
            for i, resp in enumerate(response_stream):
                print(f"resp={resp}")
                assert resp == json_stream[i]


@httprettified  # type: ignore[misc]
def test_201_request_stream() -> None:
    """Test `request_stream()` when there's no response."""
    mock_url = "http://test"
    rpc = RestClient(mock_url, "passkey", timeout=1)

    empty_streams = [
        [b"\n"],
        [],
        [b"\n", b"\r\n", b"\n"],
        [b" \n"],
        [b" "],
        [b"\t"],
    ]
    for expected_stream in empty_streams:
        # test multiple times
        for test_num in range(2):
            print(f"\niteration #{test_num}")
            HTTPretty.register_uri(
                HTTPretty.POST,
                mock_url + "/stream/no-resp/",
                body=expected_stream,
                streaming=True,
            )
            response_stream = rpc.request_stream("POST", "/stream/no-resp/", {})

            never_entered = True
            with _in_time(10, "Iterating by line is taking forever!"):
                for _ in response_stream:
                    never_entered = False
            assert never_entered

        # now w/ chunk sizes
        for chunk_size in [None, -1, 0, 1, 2, 3, 4, 8, 9, 20, 100, 1024, 32768]:
            print(f"\nchunk_size: {chunk_size}")
            HTTPretty.register_uri(
                HTTPretty.POST,
                mock_url + "/stream/no-resp/w/chunks",
                body=expected_stream,
                streaming=True,
            )
            response_stream = rpc.request_stream(
                "POST", "/stream/no-resp/w/chunks", {}, chunk_size=chunk_size
            )

            never_entered_w_chunks = True
            with _in_time(10, "Iterating by line is taking forever w/ chunks!"):
                for _ in response_stream:
                    never_entered_w_chunks = False
            assert never_entered_w_chunks


@httprettified  # type: ignore[misc]
def test_202_request_stream() -> None:
    """Test `request_stream()` where there's only one response."""
    mock_url = "http://test"
    rpc = RestClient(mock_url, "passkey", timeout=1)

    one_liners = [
        [b'"with-a-newline"\n'],
        [b'"w/o-a-newline"'],
    ]
    for expected_stream in one_liners:
        json_stream = list(_json_stream(expected_stream))

        # test multiple times
        for test_num in range(2):
            print(f"\niteration #{test_num}")
            HTTPretty.register_uri(
                HTTPretty.POST,
                mock_url + "/stream/no-resp/",
                body=expected_stream,
                streaming=True,
            )
            response_stream = rpc.request_stream("POST", "/stream/no-resp/", {})

            with _in_time(10, "Iterating by line is taking forever!"):
                for i, resp in enumerate(response_stream):
                    print(f"resp={resp}")
                    assert resp == json_stream[i]

        # now w/ chunk sizes
        for chunk_size in [None, -1, 0, 1, 2, 3, 4, 8, 9, 20, 100, 1024, 32768]:
            print(f"\nchunk_size: {chunk_size}")
            HTTPretty.register_uri(
                HTTPretty.POST,
                mock_url + "/stream/no-resp/w/chunks",
                body=expected_stream,
                streaming=True,
            )
            response_stream = rpc.request_stream(
                "POST", "/stream/no-resp/w/chunks", {}, chunk_size=chunk_size
            )

            with _in_time(10, "Iterating by line is taking forever w/ chunks!"):
                for i, resp in enumerate(response_stream):
                    print(f"resp={resp}")
                    assert resp == json_stream[i]



"""Test script for config."""

# fmt:off
# pylint: skip-file

import logging
import os
import shutil
import tempfile
import unittest

# local imports
from rest_tools.utils import from_environment

logger = logging.getLogger('config_test')


class config_test(unittest.TestCase):
    def setUp(self):
        super(config_test,self).setUp()
        self.test_dir = tempfile.mkdtemp(dir=os.getcwd())

        def cleanup():
            shutil.rmtree(self.test_dir)

        self.addCleanup(cleanup)
        environ = os.environ.copy()

        def clean_env():
            for k in list(os.environ):
                if k not in environ:
                    del os.environ[k]

        self.addCleanup(clean_env)

    def test_01_from_environment(self):
        with self.assertRaises(OSError):
            from_environment({'FOO': None})

        os.environ['FOO'] = 'bar'
        config = from_environment({'FOO': None})
        self.assertEqual(config['FOO'], 'bar')

    def test_10_from_environment(self):
        os.environ['FOO'] = 'bar'
        with self.assertRaises(ValueError):
            from_environment({'FOO': 123})

        os.environ['FOO'] = '543'
        config = from_environment({'FOO': 123})
        self.assertEqual(config['FOO'], 543)

        os.environ['FOO'] = '543.'
        config = from_environment({'FOO': 123.})
        self.assertEqual(config['FOO'], 543.)

        os.environ['FOO'] = 't'
        config = from_environment({'FOO': False})
        self.assertEqual(config['FOO'], True)


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    alltests = loader.getTestCaseNames(config_test)
    suite.addTests(loader.loadTestsFromNames(alltests,config_test))
    return suite



import logging
from typing import Any
from unittest.mock import Mock
import urllib.parse

import pytest
from requests import PreparedRequest

from rest_tools.client import DeviceGrantAuth, SavedDeviceGrantAuth
from rest_tools.client.device_client import _load_token_from_file, _save_token_to_file
from rest_tools.utils.json_util import json_encode  # isort:skip # noqa # pylint: disable=C0413


@pytest.fixture
def well_known_mock(requests_mock: Mock):
    result = {
        'token_endpoint': 'http://test/token',
        'device_authorization_endpoint': 'http://test/device',
    }

    def response(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        return json_encode(result).encode("utf-8")
    requests_mock.get("http://test/.well-known/openid-configuration", content=response)


def test_10_success(well_known_mock, requests_mock: Mock) -> None:
    device_result = {
        'device_code': 'XXXXXXXXXXXXXXXXXXXXXXXX',
        'user_code': 'XXXX-XXXX',
        'verification_uri': 'http://test/device_verification',
        'expires_in': 600,
        'interval': 0.1,
    }

    def response(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        body = urllib.parse.parse_qs(str(req.body))
        logging.debug('device request args: %r', body)
        assert body['client_id'][0] == 'client-id'
        return json_encode(device_result).encode("utf-8")
    requests_mock.post('http://test/device', content=response)

    token_result = {
        'access_token': 'XXXXXXXXXXXXX',
        'refresh_token': 'YYYYYYYYYYYYY',
        'token_type': 'bearer',
    }

    def response2(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        body = urllib.parse.parse_qs(str(req.body))
        logging.debug('token request args: %r', body)
        assert body['client_id'][0] == 'client-id'
        if body['grant_type'][0] != 'refresh_token':
            assert body['grant_type'][0] == 'urn:ietf:params:oauth:grant-type:device_code'
            assert body['device_code'][0] == device_result['device_code']
        return json_encode(token_result).encode("utf-8")
    requests_mock.post('http://test/token', content=response2)

    DeviceGrantAuth('http://test-api', 'http://test', 'client-id')


def test_20_device_unsupported(requests_mock: Mock) -> None:
    result = {
        'token_endpoint': 'http://test/token',
    }

    def response(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        return json_encode(result).encode("utf-8")
    requests_mock.get("http://test/.well-known/openid-configuration", content=response)

    with pytest.raises(RuntimeError, match='not supported'):
        DeviceGrantAuth('http://test-api', 'http://test', 'client-id')


def test_21_device_bad_client(well_known_mock, requests_mock: Mock) -> None:
    device_result = {
        'error': 'invalid_client'
    }

    def response(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        body = urllib.parse.parse_qs(str(req.body))
        logging.debug('device request args: %r', body)
        assert body['client_id'][0] == 'client-id'
        ctx.status_code = 401
        return json_encode(device_result).encode("utf-8")
    requests_mock.post('http://test/device', content=response)

    with pytest.raises(RuntimeError, match='invalid_client'):
        DeviceGrantAuth('http://test-api', 'http://test', 'client-id')


def test_22_device_code_timeout(well_known_mock, requests_mock: Mock) -> None:
    device_result = {
        'device_code': 'XXXXXXXXXXXXXXXXXXXXXXXX',
        'user_code': 'XXXX-XXXX',
        'verification_uri': 'http://test/device_verification',
        'expires_in': 600,
        'interval': 0.1,
    }

    def response(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        body = urllib.parse.parse_qs(str(req.body))
        logging.debug('device request args: %r', body)
        assert body['client_id'][0] == 'client-id'
        return json_encode(device_result).encode("utf-8")
    requests_mock.post('http://test/device', content=response)

    token_result = {
        'error': 'expired_token'
    }

    def response2(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        body = urllib.parse.parse_qs(str(req.body))
        logging.debug('token request args: %r', body)
        assert body['client_id'][0] == 'client-id'
        if body['grant_type'][0] != 'refresh_token':
            assert body['grant_type'][0] == 'urn:ietf:params:oauth:grant-type:device_code'
            assert body['device_code'][0] == device_result['device_code']
        ctx.status_code = 400
        return json_encode(token_result).encode("utf-8")
    requests_mock.post('http://test/token', content=response2)

    with pytest.raises(RuntimeError, match='expired_token'):
        DeviceGrantAuth('http://test-api', 'http://test', 'client-id')


def test_23_device_code_denied(well_known_mock, requests_mock: Mock) -> None:
    device_result = {
        'device_code': 'XXXXXXXXXXXXXXXXXXXXXXXX',
        'user_code': 'XXXX-XXXX',
        'verification_uri': 'http://test/device_verification',
        'expires_in': 600,
        'interval': 0.1,
    }

    def response(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        body = urllib.parse.parse_qs(str(req.body))
        logging.debug('device request args: %r', body)
        assert body['client_id'][0] == 'client-id'
        return json_encode(device_result).encode("utf-8")
    requests_mock.post('http://test/device', content=response)

    token_result = {
        'error': 'access_denied'
    }

    def response2(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        body = urllib.parse.parse_qs(str(req.body))
        logging.debug('token request args: %r', body)
        assert body['client_id'][0] == 'client-id'
        if body['grant_type'][0] != 'refresh_token':
            assert body['grant_type'][0] == 'urn:ietf:params:oauth:grant-type:device_code'
            assert body['device_code'][0] == device_result['device_code']
        ctx.status_code = 400
        return json_encode(token_result).encode("utf-8")
    requests_mock.post('http://test/token', content=response2)

    with pytest.raises(RuntimeError, match='access_denied'):
        DeviceGrantAuth('http://test-api', 'http://test', 'client-id')


def test_100_saved_no_token(well_known_mock, requests_mock, tmp_path) -> None:
    device_result = {
        'device_code': 'XXXXXXXXXXXXXXXXXXXXXXXX',
        'user_code': 'XXXX-XXXX',
        'verification_uri': 'http://test/device_verification',
        'expires_in': 600,
        'interval': 0.1,
    }

    def response(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        body = urllib.parse.parse_qs(str(req.body))
        logging.debug('device request args: %r', body)
        assert body['client_id'][0] == 'client-id'
        return json_encode(device_result).encode("utf-8")
    requests_mock.post('http://test/device', content=response)

    token_result = {
        'access_token': 'XXXXXXXXXXXXX',
        'refresh_token': 'YYYYYYYYYYYYY',
        'token_type': 'bearer',
    }

    def response2(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        body = urllib.parse.parse_qs(str(req.body))
        logging.debug('token request args: %r', body)
        assert body['client_id'][0] == 'client-id'
        if body['grant_type'][0] != 'refresh_token':
            assert body['grant_type'][0] == 'urn:ietf:params:oauth:grant-type:device_code'
            assert body['device_code'][0] == device_result['device_code']
        return json_encode(token_result).encode("utf-8")
    requests_mock.post('http://test/token', content=response2)

    filepath = tmp_path / 'foo'
    SavedDeviceGrantAuth('http://test-api', 'http://test', filename=str(filepath), client_id='client-id')

    assert filepath.read_text() == token_result['refresh_token']


def test_101_load_save(tmp_path) -> None:
    filepath = tmp_path / 'foo'
    token = '12345'

    _save_token_to_file(filepath, token)
    assert token == _load_token_from_file(filepath)


def test_110_saved_good_token(well_known_mock, requests_mock, tmp_path) -> None:
    device_result = {
        'device_code': 'XXXXXXXXXXXXXXXXXXXXXXXX',
        'user_code': 'XXXX-XXXX',
        'verification_uri': 'http://test/device_verification',
        'expires_in': 600,
        'interval': 0.1,
    }

    def response(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        raise NotImplementedError()
    mock_device = requests_mock.post('http://test/device', content=response)

    token_result = {
        'access_token': 'XXXXXXXXXXXXX',
        'refresh_token': 'YYYYYYYYYYYYY',
        'token_type': 'bearer',
    }

    def response2(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        body = urllib.parse.parse_qs(str(req.body))
        logging.debug('token request args: %r', body)
        assert body['client_id'][0] == 'client-id'
        if body['grant_type'][0] != 'refresh_token':
            assert body['grant_type'][0] == 'urn:ietf:params:oauth:grant-type:device_code'
            assert body['device_code'][0] == device_result['device_code']
        return json_encode(token_result).encode("utf-8")
    mock_refresh = requests_mock.post('http://test/token', content=response2)

    filepath = tmp_path / 'foo'
    _save_token_to_file(filepath, 'badtoken')
    SavedDeviceGrantAuth('http://test-api', 'http://test', filename=str(filepath), client_id='client-id')

    assert filepath.read_text() == token_result['refresh_token']
    assert mock_device.call_count == 0
    assert mock_refresh.call_count == 1


def test_111_saved_expired_token(well_known_mock, requests_mock, tmp_path) -> None:
    device_result = {
        'device_code': 'XXXXXXXXXXXXXXXXXXXXXXXX',
        'user_code': 'XXXX-XXXX',
        'verification_uri': 'http://test/device_verification',
        'expires_in': 600,
        'interval': 0.1,
    }

    def response(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        body = urllib.parse.parse_qs(str(req.body))
        logging.debug('device request args: %r', body)
        assert body['client_id'][0] == 'client-id'
        return json_encode(device_result).encode("utf-8")
    mock_device = requests_mock.post('http://test/device', content=response)

    token_result = {
        'access_token': 'XXXXXXXXXXXXX',
        'refresh_token': 'YYYYYYYYYYYYY',
        'token_type': 'bearer',
    }

    def response2(req: PreparedRequest, ctx: Any) -> bytes:  # pylint: disable=W0613
        assert req.body is not None
        body = urllib.parse.parse_qs(str(req.body))
        logging.debug('token request args: %r', body)
        assert body['client_id'][0] == 'client-id'
        if body['grant_type'][0] == 'refresh_token':
            assert body['refresh_token'][0] != 'badtoken'
        else:
            assert body['grant_type'][0] == 'urn:ietf:params:oauth:grant-type:device_code'
            assert body['device_code'][0] == device_result['device_code']
        return json_encode(token_result).encode("utf-8")
    mock_refresh = requests_mock.post('http://test/token', content=response2)

    filepath = tmp_path / 'foo'
    _save_token_to_file(filepath, 'badtoken')
    SavedDeviceGrantAuth('http://test-api', 'http://test', filename=str(filepath), client_id='client-id')

    assert filepath.read_text() == token_result['refresh_token']
    assert mock_device.call_count == 1
    assert mock_refresh.call_count == 3



import argparse

from rest_tools.client import ClientCredentialsAuth


def get_token(address, client_id, client_secret):
    rest_client = ClientCredentialsAuth('', token_url=address, client_id=client_id, client_secret=client_secret)
    # could use `rest_client` as a `RestClient`, but we're only interested in the token
    return rest_client.make_access_token()


def main():
    parser = argparse.ArgumentParser(description='Get an OAuth2 token via client credentials')
    parser.add_argument('--address', default='https://keycloak.icecube.wisc.edu/auth/realms/IceCube', help='OAuth2 server address')
    parser.add_argument('client_id', help='client id')
    parser.add_argument('client_secret', help='client secret')

    args = parser.parse_args()
    kwargs = vars(args)
    print('access token:', get_token(**kwargs))


if __name__ == '__main__':
    main()



"""Test server/arghandler.py."""

import argparse
import json
import re
import sys
from typing import Any, Dict, List, Tuple, Union, cast
from unittest.mock import Mock

import pytest
import requests
import tornado.web
from tornado import httputil
from wipac_dev_tools import strtobool

from rest_tools.server.arghandler import ArgumentHandler, ArgumentSource
from rest_tools.server.handler import RestHandler

# pylint: disable=W0212,W0621

# these tests are only for 3.9+
if sys.version_info < (3, 9):
    pytest.skip("only for 3.9+", allow_module_level=True)  # type: ignore[var-annotated]

QUERY_ARGUMENTS = "query-arguments"
JSON_BODY_ARGUMENTS = "json-body-arguments"


def setup_argument_handler(
    argument_source: str,
    args: Union[Dict[str, Any], List[Tuple[Any, Any]]],
) -> ArgumentHandler:
    """Load data and return `ArgumentHandler` instance."""
    if argument_source == QUERY_ARGUMENTS:
        req = requests.Request(url="https://foo.aq/all", params=args).prepare()
        print("\n request:")
        print(req.url)
        print()
        rest_handler = RestHandler(
            application=Mock(),
            request=httputil.HTTPServerRequest(
                uri=req.url,
                headers=httputil.HTTPHeaders(req.headers),
            ),
        )
        return ArgumentHandler(ArgumentSource.QUERY_ARGUMENTS, rest_handler)

    elif argument_source == JSON_BODY_ARGUMENTS:
        req = requests.Request(url="https://foo.aq/all", json=args).prepare()
        print("\n request:")
        print(req.url)
        print(req.body)
        print(req.headers)
        print()
        rest_handler = RestHandler(
            application=Mock(),
            request=httputil.HTTPServerRequest(
                uri=req.url,
                body=cast(bytes, req.body),
                headers=httputil.HTTPHeaders(req.headers),
            ),
        )
        rest_handler.request._parse_body()  # normally called when request REST method starts
        print("\n rest-handler:")
        print(rest_handler.request.body)
        print(json.loads(rest_handler.request.body))
        print()
        return ArgumentHandler(ArgumentSource.JSON_BODY_ARGUMENTS, rest_handler)

    else:
        raise ValueError(f"Invalid argument_source: {argument_source}")


########################################################################################


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_100__defaults(argument_source: str) -> None:
    """Test `argument_source` arguments with default."""
    default: Any

    for default in [None, "a-string", 100, 50.5]:
        print(default)
        arghand = setup_argument_handler(argument_source, {})
        arghand.add_argument("myarg", default=default)
        args = arghand.parse_args()
        assert default == args.myarg

    # w/ typing
    for default in ["a-string", 100, 50.5]:
        print(default)
        arghand = setup_argument_handler(argument_source, {})
        arghand.add_argument("myarg", default=default, type=type(default))
        args = arghand.parse_args()
        assert default == args.myarg


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_110__no_default_no_typing(argument_source: str) -> None:
    """Test `argument_source` arguments."""
    args: Dict[str, Any] = {
        "foo": "-10",
        "bar": "True",
        "bat": "2.5",
        "baz": "Toyota Corolla",
        "boo": "False",
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args.update(
            {
                "dicto": {"abc": 123},
                "listo": [1, 2, 3],
                "falso": False,
                "truo": True,
                "compoundo": [{"apple": True}, {"banana": 951}, {"cucumber": False}],
            }
        )

    # set up ArgumentHandler
    arghand = setup_argument_handler(argument_source, args)

    for arg, _ in args.items():
        print()
        print(arg)
        arghand.add_argument(arg)
    outargs = arghand.parse_args()

    # grab each
    for arg, val in args.items():
        print()
        print(arg)
        print(val)
        assert val == getattr(outargs, arg.replace("-", "_"))


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_111__no_default_with_typing(argument_source: str) -> None:
    """Test `request.arguments` arguments."""

    def custom_type(_: Any) -> Any:
        return "this could be anything but it's just a string"

    args: Dict[str, Any] = {
        "foo": ("-10", int),
        #
        "bar": ("True", bool),
        "boo": ("False", bool),
        "ghi": ("no", bool),
        "jkl": ("1", bool),
        "mno": ("0", bool),
        #
        "bat": ("2.5", float),
        "abc": ("2", float),
        #
        "baz": ("Toyota Corolla", str),
        "def": ("1000", str),
        #
        "zzz": ("yup", custom_type),
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args.update(
            {
                "dicto": ({"abc": 123}, dict),
                "dicto-as-list": ({"def": 456}, list),
                "listo": ([1, 2, 3], list),
                "falso": (False, bool),
                "truo": (True, bool),
                "compoundo": (
                    [{"apple": True}, {"banana": 951}, {"cucumber": False}],
                    list,
                ),
                "sure": (
                    [{"apple": True}, {"banana": 951}, {"cucumber": False}],
                    custom_type,
                ),
            }
        )

    # set up ArgumentHandler
    arghand = setup_argument_handler(
        argument_source, {arg: val for arg, (val, _) in args.items()}
    )
    for arg, (_, typ) in args.items():
        print()
        print(arg)
        arghand.add_argument(arg, type=typ)
    outargs = arghand.parse_args()

    # grab each
    for arg, (val, typ) in args.items():
        print()
        print(arg)
        print(val)
        print(typ)
        if typ == bool and isinstance(val, str):
            assert getattr(outargs, arg.replace("-", "_")) == strtobool(val)
        else:
            assert getattr(outargs, arg.replace("-", "_")) == typ(val)


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_112__no_default_with_typing__error(argument_source: str) -> None:
    """Test `argument_source` arguments."""

    def custom_type(val: Any) -> Any:
        raise ValueError("x")

    args: Dict[str, Any] = {
        "foo": ("hank", int),
        "bat": ("2.5", int),
        "baz": ("9e-33", int),
        #
        "bar": ("idk", bool),
        "boo": ("2", bool),
        #
        "abc": ("222.111.333", float),
        #
        "zzz": ("yup", custom_type),
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args.update(
            {
                "dicto": ({"abc": 123}, int),
                "listo": ([1, 2, 3], int),
                "compoundo": (
                    [{"apple": True}, {"banana": 951}, {"cucumber": False}],
                    object,
                ),
                "nope": (
                    [{"apple": True}, {"banana": 951}, {"cucumber": False}],
                    custom_type,
                ),
            }
        )

    for arg, (val, typ) in args.items():
        print()
        print(arg)
        print(val)
        print(typ)
        arghand = setup_argument_handler(argument_source, {arg: val})
        arghand.add_argument(arg, type=typ)
        with pytest.raises(tornado.web.HTTPError) as e:
            arghand.parse_args()
        assert str(e.value) == f"HTTP 400: argument {arg}: invalid type"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_120__missing_argument(argument_source: str) -> None:
    """Test `argument_source` arguments error case."""

    # set up ArgumentHandler
    arghand = setup_argument_handler(argument_source, {})
    arghand.add_argument("reqd")

    # Missing Required Argument
    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()
    assert str(e.value) == "HTTP 400: the following arguments are required: reqd"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_121__missing_argument(argument_source: str) -> None:
    """Test `argument_source` arguments error case."""

    # set up ArgumentHandler
    arghand = setup_argument_handler(argument_source, {"foo": "val"})
    arghand.add_argument("reqd")
    arghand.add_argument("foo")
    arghand.add_argument("bar")

    # Missing Required Argument
    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()
    assert str(e.value) == "HTTP 400: the following arguments are required: reqd, bar"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_130__duplicates(argument_source: str) -> None:
    """Test `argument_source` arguments with duplicate keys."""

    args = [
        ("foo", "22"),
        ("foo", "44"),
        ("foo", "66"),
        ("foo", "88"),
        ("bar", "abc"),
        ("baz", "hello world!"),
        ("baz", "hello mars!"),
    ]
    if argument_source == JSON_BODY_ARGUMENTS:
        pass  # no special data since duplicates not allowed for json

    # set up ArgumentHandler
    arghand = setup_argument_handler(argument_source, args)
    arghand.add_argument("foo", type=int, nargs="*")
    arghand.add_argument("baz", type=str, nargs="*")
    arghand.add_argument("bar")

    # duplicates not allowed for json
    if argument_source == JSON_BODY_ARGUMENTS:
        with pytest.raises(tornado.web.HTTPError) as e:
            arghand.parse_args()
        assert str(e.value) == "HTTP 400: JSON-encoded requests body must be a 'dict'"
    # but they are for query args!
    else:
        outargs = arghand.parse_args()
        print(outargs)
        # grab each
        assert outargs.foo == [22, 44, 66, 88]
        assert outargs.baz == ["hello world!", "hello mars!"]
        assert outargs.bar == "abc"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_140__extra_argument(argument_source: str) -> None:
    """Test `argument_source` arguments error case."""

    # set up ArgumentHandler
    args: Dict[str, Any] = {
        "foo": "val",
        "reqd": "2",
        "xtra": 1,
        "another": True,
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args["another"] = [{"apple": True}, {"banana": 951}, {"cucumber": False}]

    arghand = setup_argument_handler(argument_source, args)
    arghand.add_argument("reqd")
    arghand.add_argument("foo")

    # Missing Required Argument
    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()
    assert str(e.value) == "HTTP 400: unrecognized arguments: xtra, another"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_141__extra_argument_with_duplicates(argument_source: str) -> None:
    """Test `argument_source` arguments error case."""

    # set up ArgumentHandler
    args = [
        ("foo", "val"),
        ("reqd", "2"),
        ("xtra", 1),
        ("another", True),
        ("another", False),
        ("another", "who knows"),
    ]
    if argument_source == JSON_BODY_ARGUMENTS:
        pass  # no special data since duplicates not allowed for json

    arghand = setup_argument_handler(argument_source, args)
    arghand.add_argument("reqd")
    arghand.add_argument("foo")

    # Missing Required Argument...

    # duplicates not allowed for json
    if argument_source == JSON_BODY_ARGUMENTS:
        with pytest.raises(tornado.web.HTTPError) as e:
            arghand.parse_args()
        assert str(e.value) == "HTTP 400: JSON-encoded requests body must be a 'dict'"
    # but they are for query args!
    else:
        with pytest.raises(tornado.web.HTTPError) as e:
            arghand.parse_args()
        assert str(e.value) == "HTTP 400: unrecognized arguments: xtra, another"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_200__argparse_dest(argument_source: str) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: Dict[str, Any] = {
        "old_name": "-10",
        "bar": "True",
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args.update(
            {
                "dicto": {"abc": 123},
                "listo": [1, 2, 3],
            }
        )

    # set up ArgumentHandler
    arghand = setup_argument_handler(argument_source, args)

    for arg, _ in args.items():
        print()
        print(arg)
        if arg == "old_name":
            arghand.add_argument(arg, dest="new_name")
        else:
            arghand.add_argument(arg)
    outargs = arghand.parse_args()

    # grab each
    for arg, val in args.items():
        print()
        print(arg)
        print(val)
        if arg == "old_name":
            assert val == getattr(outargs, "new_name")
        else:
            assert val == getattr(outargs, arg.replace("-", "_"))


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_210__argparse_choices(argument_source: str) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: Dict[str, Any] = {
        "pick_it": "paper",
        "bar": "True",
    }
    choices: List = ["rock", "paper", "scissors"]
    if argument_source == JSON_BODY_ARGUMENTS:
        args.update(
            {
                "pick_it": {"abc": 123},
                "listo": [1, 2, 3],
            }
        )
        choices = [{"abc": 123}, {"def": 456}, {"ghi": 789}]

    # set up ArgumentHandler
    arghand = setup_argument_handler(argument_source, args)

    for arg, _ in args.items():
        print()
        print(arg)
        if arg == "pick_it":
            arghand.add_argument(arg, choices=choices)
        else:
            arghand.add_argument(arg)
    outargs = arghand.parse_args()

    # grab each
    for arg, val in args.items():
        print()
        print(arg)
        print(val)
        assert val == getattr(outargs, arg.replace("-", "_"))


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_211__argparse_choices__error(argument_source: str) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: Dict[str, Any] = {
        "pick_it": "paper",
        "bar": "True",
    }
    choices: List = ["rock", "paper", "scissors"]
    if argument_source == JSON_BODY_ARGUMENTS:
        args.update(
            {
                "pick_it": {"abc": 123},
                "listo": [1, 2, 3],
            }
        )
        choices = [{"abc": 123}, {"def": 456}, {"ghi": 789}]

    # set up ArgumentHandler
    arghand = setup_argument_handler(
        argument_source,
        {k: v if k != "pick_it" else "hank" for k, v in args.items()},
    )

    for arg, _ in args.items():
        print()
        print(arg)
        if arg == "pick_it":
            arghand.add_argument(arg, choices=choices)
        else:
            arghand.add_argument(arg)

    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()

    # For Python 3.13 and later: join stringified representations directly
    expected_message1 = (
        f"HTTP 400: argument pick_it: invalid choice: 'hank' "
        f"(choose from {', '.join(str(c) for c in choices)})"
    )
    # For Python 3.9–3.12: join repr representations directly
    expected_message2 = (
        f"HTTP 400: argument pick_it: invalid choice: 'hank' "
        f"(choose from {', '.join(repr(c) for c in choices)})"
    )
    assert (
        str(e.value) == expected_message1 or str(e.value) == expected_message2
    ), "Error does not match expected values"


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_220__argparse_nargs(argument_source: str) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    test_130__duplicates(argument_source)


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
@pytest.mark.parametrize(
    "exc",
    [TypeError, ValueError, argparse.ArgumentError],
)
def test_230__argparse_custom_validation__supported_builtins__error(
    argument_source: str, exc: Exception
) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: Dict[str, Any] = {
        "foo": "True",
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args = {
            "foo": [1, 2, 3],
        }

    # set up ArgumentHandler
    arghand = setup_argument_handler(
        argument_source,
        args,
    )

    def _error_it(_: Any, exc: Exception):
        raise exc

    for arg, _ in args.items():
        print()
        print(arg)
        arghand.add_argument(
            arg,
            type=lambda x: _error_it(
                x,
                exc("it's a bad value but you won't see this message anyway..."),  # type: ignore
            ),
        )

    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()

    assert str(e.value) == "HTTP 400: argument foo: invalid type"


class MyError(Exception):
    """Used below."""


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
@pytest.mark.parametrize(
    "exc",
    [RuntimeError, IndexError, MyError],
)
def test_232__argparse_custom_validation__unsupported_errors__error(
    argument_source: str, exc: Exception
) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: Dict[str, Any] = {
        "foo": "True",
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args = {
            "foo": [1, 2, 3],
        }

    # set up ArgumentHandler
    arghand = setup_argument_handler(
        argument_source,
        args,
    )

    def _error_it(_: Any, exc: Exception):
        raise exc

    for arg, _ in args.items():
        print()
        print(arg)
        arghand.add_argument(
            arg,
            type=lambda x: _error_it(
                x,
                exc("something went wrong but not in an unexpected way, not-validation"),  # type: ignore
            ),
        )

    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()

    assert re.fullmatch(
        r"HTTP 400: Unknown argument-handling error \(\d+\.\d+\)", str(e.value)
    )


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
def test_234__argparse_custom_validation__argumenttypeerror__error(
    argument_source: str,
) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: Dict[str, Any] = {
        "foo": "True",
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args = {
            "foo": [1, 2, 3],
        }

    # set up ArgumentHandler
    arghand = setup_argument_handler(
        argument_source,
        args,
    )

    def _error_it(_: Any):
        raise argparse.ArgumentTypeError("it's a bad value and you *will* see this!")

    for arg, _ in args.items():
        print()
        print(arg)
        arghand.add_argument(
            arg,
            type=_error_it,
        )

    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()

    assert (
        str(e.value)
        == "HTTP 400: argument foo: it's a bad value and you *will* see this!"
    )


@pytest.mark.parametrize(
    "argument_source",
    [QUERY_ARGUMENTS, JSON_BODY_ARGUMENTS],
)
@pytest.mark.parametrize(
    "exc",
    [  # all the exceptions!
        TypeError,
        ValueError,
        argparse.ArgumentError,
        argparse.ArgumentTypeError,
        RuntimeError,
        IndexError,
        MyError,
    ],
)
def test_236__argparse_custom_validation__validator_no_param__error(
    argument_source: str,
    exc: Exception,
) -> None:
    """Test `argument_source` arguments using argparse's advanced options."""
    args: Dict[str, Any] = {
        "foo": "True",
    }
    if argument_source == JSON_BODY_ARGUMENTS:
        args = {
            "foo": [1, 2, 3],
        }

    # set up ArgumentHandler
    arghand = setup_argument_handler(
        argument_source,
        args,
    )

    def _error_it__no_param():
        raise exc("it's a bad value and you *will* see this!")  # type: ignore

    for arg, _ in args.items():
        print()
        print(arg)
        arghand.add_argument(
            arg,
            type=_error_it__no_param,
            # NOTE: ^^^ because this takes no arguments (isn't a func/lambda),
            #       argparse treats it like any other error. why? idk :/
        )

    with pytest.raises(tornado.web.HTTPError) as e:
        arghand.parse_args()

    assert str(e.value) == "HTTP 400: argument foo: invalid type"



"""A simple REST json client using `requests`_ for the http connection.

.. _requests: http://docs.python-requests.org

The REST protocol is built on http(s), with the body containing
a json-encoded dictionary as necessary.
"""

import logging
from typing import Any, Callable, Optional, Union

import jwt
import requests
from requests.auth import HTTPBasicAuth

from .client import RestClient
from ..utils.auth import OpenIDAuth


# fmt:off


class OpenIDRestClient(RestClient):
    """A REST client that can handle token refresh using OpenID .well-known
    auto-discovery.

    Args:
        address (str): base address of REST API
        token_url (str): base address of token service
        client_id (str): client id
        client_secret (str): client secret (optional - required to generate new refresh token)
        refresh_token (str): initial refresh token (optional)
        update_func (callable): a function that gets called when the access and refresh tokens are updated (optional)
        timeout (int): request timeout (optional)
        retries (int): number of retries to attempt (optional)
    """

    def __init__(
        self,
        address: str,
        token_url: str,
        refresh_token: Union[str, bytes],
        client_id: str,
        client_secret: Optional[str] = None,
        update_func: Optional[
            Callable[[Union[str, bytes], Optional[Union[str, bytes]]], None]
        ] = None,
        **kwargs: Any,
    ) -> None:
        self.auth = OpenIDAuth(token_url)
        self.refresh_token = refresh_token
        self.client_id = client_id
        self.client_secret = client_secret
        self.update_func = update_func

        super().__init__(
            address,
            logger=kwargs.pop('logger', logging.getLogger('OpenIDRestClient')),
            token=self._openid_token,
            **kwargs,
        )

        # initial call to verify things work
        self._openid_token()

    def _get_scopes(self) -> str:
        try:
            return jwt.decode(self.refresh_token, options={'verify_signature': False}).get('scope', '')
        except jwt.exceptions.DecodeError:
            # in case the token is opaque
            return ''

    def _openid_token(self) -> str:
        if not self.auth.token_url:
            self.auth._refresh_keys()

        # try the refresh token
        scopes = self._get_scopes()
        self.logger.debug('refreshing the token. scopes: %s', scopes)
        args = {
            'grant_type': 'refresh_token',
            'refresh_token': self.refresh_token,
            'client_id': self.client_id,
            'scope': scopes,
        }
        kwargs = {}
        if self.client_secret:
            kwargs['auth'] = HTTPBasicAuth(self.client_id, self.client_secret)

        try:
            r = requests.post(self.auth.token_url, data=args, **kwargs)  # type: ignore[arg-type]
            r.raise_for_status()
            req = r.json()
        except requests.exceptions.HTTPError as exc:
            self.logger.debug('%r', exc.response.text)
            try:
                req = exc.response.json()
            except Exception:
                req = {}
            error = req.get('error', '')
            raise Exception(f'Token request failed: {error}') from exc
        else:
            self.logger.debug('OpenID token refreshed')
            access_token = req['access_token']
            if 'refresh_token' in req:
                self.refresh_token = req['refresh_token']
                if access_token and self.update_func:
                    self.update_func(access_token, self.refresh_token)
            return access_token


class RegisterOpenIDClient:
    def __init__(self, address: str, client_name: str, scopes: Union[list[str], None] = None):
        self.address = address
        self.client_name = client_name
        self.scopes = scopes if scopes else []
        self._registration_path = None
        self._registration_token = None

    async def register_client(self):
        rc = RestClient(self.address, retries=0)
        meta = {
            'grant_types': [
                'authorization_code',
                'urn:ietf:params:oauth:grant-type:device_code',
            ],
            'client_name': self.client_name,
            'scope': ' '.join(self.scopes),
        }
        ret = await rc.request('POST', '/client', meta)
        self._registration_path = ret['registration_client_uri'][len(self.address):]
        self._registration_token = ret['registration_access_token']

        client_id = ret['client_id']
        client_secret = ret['client_secret']
        return (client_id, client_secret)

    async def delete_client(self, *args, **kwargs):
        if self._registration_path and self._registration_token:
            rc2 = RestClient(self.address, token=self._registration_token, retries=0)
            await rc2.request('DELETE', self._registration_path)

    __aenter__ = register_client
    __aexit__ = delete_client



"""Test server.stats."""

# fmt:off
# pylint: skip-file

import time

import pytest

# local imports
from rest_tools.server import stats


@pytest.fixture
def random_half(mocker):
    def new_random():
        return 0.5
    mocker.patch('random.random', new_random)


def test_stats_empty():
    s = stats.RouteStats()
    assert not s.is_overloaded()


def test_stats_basic(random_half):
    s = stats.RouteStats(window_size=10, timeout=10)
    for i in range(10):
        s.append(i)
    assert not s.is_overloaded()

    s.clear()
    for i in range(100,110):
        s.append(i)
    assert s.is_overloaded()
    assert 205 < s.get_backoff_time() < 215


def test_stats_expiring(random_half):
    s = stats.RouteStats(window_size=100, window_time=5, timeout=10)

    for i in range(100,105):
        s.append(i)
    time.sleep(5)
    for i in range(10):
        s.append(i)

    assert not s.is_overloaded()



"""Route stats."""


# fmt:off

import logging
import random
import statistics
import time
from collections import deque

LOGGER = logging.getLogger(__name__)


class RouteStats:
    """
    Route tracking and statistics.

    Keeps track of the last N calls to a route,
    and presents stats on their performance.

    Args:
        window_size (int): number of past calls to track
        window_time (int): number of seconds to keep track of past calls
        timeout (int): seconds before a request is considered "over time"
    """
    def __init__(self, window_size=1000, window_time=3600, timeout=30):
        self.data = deque(maxlen=window_size)
        self.times = deque(maxlen=window_size)
        self.window_size = window_size
        self.window_time = window_time
        self.timeout = timeout

    def __len__(self):
        return len(self.data)

    def append(self, call_time):
        self.data.append(call_time)
        self.times.append(time.time())

    def clear(self):
        self.data.clear()
        self.times.clear()

    def is_overloaded(self):
        # check window time
        window_cutoff = time.time()-self.window_time
        i = 0
        for i,t in enumerate(self.times):
            if t >= window_cutoff:
                break
        if i > 0:
            LOGGER.debug('routestats: removing %d entries due to age', i)
            self.data = deque((self.data[j] for j in range(i,len(self.data))), maxlen=self.window_size)
            self.times = deque((self.times[j] for j in range(i,len(self.times))), maxlen=self.window_size)

        # check if we have enough data to form stats
        if len(self.data) < 4:
            return False

        # now check stats
        median = 0
        try:
            stats = statistics.quantiles(self.data)
            LOGGER.debug('routestats: %r',stats)
            if stats[1] >= self.timeout or stats[2] >= 2*self.timeout:
                median = stats[1]
        except AttributeError:
            med = statistics.median(self.data)
            LOGGER.debug('routestats: %r',med)
            if med >= self.timeout:
                median = med
        return median > 0 and random.random()*median >= self.timeout

    def get_backoff_time(self):
        if len(self.data) < 4:
            return 1
        return int(statistics.median(self.data)*2)



"""device_client.py"""

import io
import logging
import time
from itertools import zip_longest
from pathlib import Path
from typing import Any, Dict, List, Optional

import qrcode  # type: ignore[import]
import requests
from requests.auth import HTTPBasicAuth

from .openid_client import OpenIDRestClient
from ..utils.auth import OpenIDAuth
from ..utils.pkce import PKCEMixin


# fmt:off


def _print_qrcode(req: Dict[str, str]) -> None:
    if 'verification_uri_complete' not in req:
        req['verification_uri_complete'] = req['verification_uri']+'?user_code='+req['user_code']

    qr = qrcode.QRCode(border=2)
    qr.add_data(req['verification_uri_complete'])
    f = io.StringIO()
    qr.print_ascii(out=f)
    f.seek(0)
    code = f.read().rstrip().split('\n')
    code_width = max(len(c) for c in code)
    text_width = max(len(req['verification_uri']), 30)
    box_width = 2 + text_width + code_width

    qrtext = f'''
To complete authorization,
scan the QR code or, using
a browser on another device,
visit:
{req["verification_uri"]}

And enter the code:
{req["user_code"]}
'''.split('\n')

    if box_width < 78:
        print('+', '-' * box_width, '+', sep='')
        for text,qrdata in zip_longest(qrtext, code, fillvalue=''):
            print('|  ', f'{text:<{text_width}}', f'{qrdata:{code_width}}', '|', sep='')
        print('+', '-' * box_width, '+', sep='')
    else:
        box_width = max(text_width+4, code_width)
        print('+', '-' * box_width, '+', sep='')
        for text in qrtext:
            print('|  ', f'{text:<{box_width-4}}', '  |', sep='')
        for qrdata in code:
            print('|', f'{qrdata:<{box_width}}', '|', sep='')
        print('+', '-' * box_width, '+', sep='')


class CommonDeviceGrant(PKCEMixin):
    def perform_device_grant(
        self,
        logger: logging.Logger,
        device_url: str,
        token_url: str,
        client_id: str,
        client_secret: Optional[str] = None,
        scopes: Optional[List[str]] = None,
    ) -> str:
        args = {
            'client_id': client_id,
            'scope': 'offline_access ' + (' '.join(scopes) if scopes else ''),
        }
        logger.debug('device grant args: %r', args)
        kwargs = {}
        if client_secret:
            kwargs['auth'] = HTTPBasicAuth(client_id, client_secret)
        else:
            code_challenge = self.create_pkce_challenge()
            args['code_challenge'] = code_challenge
            args['code_challenge_method'] = 'S256'

        try:
            r = requests.post(device_url, data=args, **kwargs)  # type: ignore[arg-type]
            r.raise_for_status()
            req = r.json()
        except requests.exceptions.HTTPError as exc:
            logger.debug('%r', exc.response.text)
            try:
                req = exc.response.json()
            except Exception:
                req = {}
            error = req.get('error', '')
            raise RuntimeError(f'Device authorization failed: {error}') from exc
        except Exception as exc:
            raise RuntimeError('Device authorization failed') from exc

        logger.debug('Device auth in progress')

        _print_qrcode(req)

        args = {
            'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
            'device_code': req['device_code'],
            'client_id': client_id,
        }
        kwargs = {}
        if client_secret:
            kwargs['auth'] = HTTPBasicAuth(client_id, client_secret)
        else:
            args['code_verifier'] = self.get_pkce_verifier(code_challenge)

        sleep_time = int(req.get('interval', 5))
        while True:
            time.sleep(sleep_time)
            try:
                r = requests.post(token_url, data=args, **kwargs)  # type: ignore[arg-type]
                r.raise_for_status()
                req = r.json()
            except requests.exceptions.HTTPError as exc:
                logger.debug('%r', exc.response.text)
                try:
                    req = exc.response.json()
                except Exception:
                    req = {}
                error = req.get('error', '')
                if error == 'authorization_pending':
                    continue
                elif error == 'slow_down':
                    sleep_time += 5
                    continue
                raise RuntimeError(f'Device authorization failed: {error}') from exc
            except Exception as exc:
                raise RuntimeError('Device authorization failed') from exc
            break

        logger.debug('device grant response: %r', req)
        return req['refresh_token']


def DeviceGrantAuth(
    address: str,
    token_url: str,
    client_id: str,
    client_secret: Optional[str] = None,
    scopes: Optional[List[str]] = None,
    **kwargs: Any,
) -> OpenIDRestClient:
    """A REST client that can handle OpenID and the OAuth2 Device Client flow.

    Args:
        address (str): base address of REST API
        token_url (str): base address of token service
        client_id (str): client id
        client_secret (str): client secret (optional - required for confidential clients)
        scopes (list): token scope list (optional)
        timeout (int): request timeout (optional)
        retries (int): number of retries to attempt (optional)
    """
    logger = kwargs.pop('logger', logging.getLogger('DeviceGrantAuth'))

    auth = OpenIDAuth(token_url)
    if 'device_authorization_endpoint' not in auth.provider_info:
        raise RuntimeError('Device grant not supported by server')
    endpoint: str = auth.provider_info['device_authorization_endpoint']  # type: ignore

    device = CommonDeviceGrant()
    refresh_token = device.perform_device_grant(
        logger, endpoint, auth.token_url, client_id, client_secret, scopes
    )

    return OpenIDRestClient(
        address=address,
        token_url=token_url,
        client_id=client_id,
        client_secret=client_secret,
        refresh_token=refresh_token,
        logger=logger,
        **kwargs,
    )


def _load_token_from_file(filepath: Path) -> Optional[str]:
    if filepath.exists():
        return filepath.read_text()
    return None


def _save_token_to_file(filepath: Path, token: str) -> None:
    filepath.write_text(token)


def SavedDeviceGrantAuth(
    address: str,
    token_url: str,
    filename: str,
    client_id: str,
    client_secret: Optional[str] = None,
    scopes: Optional[List[str]] = None,
    **kwargs: Any,
) -> OpenIDRestClient:
    """
    A REST client that can handle OpenID and the OAuth2 Device Client flow,
    and saves a refresh token to a file for quick access.

    Args:
        address (str): base address of REST API
        token_url (str): base address of token service
        filename (str): name of file to save/load refresh token
        client_id (str): client id
        client_secret (str): client secret (optional - required for confidential clients)
        scopes (list): token scope list (optional)
        timeout (int): request timeout (optional)
        retries (int): number of retries to attempt (optional)
    """
    logger = kwargs.pop('logger', logging.getLogger('SavedDeviceGrantAuth'))
    filepath = Path(filename)

    def update_func(access, refresh):
        _save_token_to_file(filepath, refresh)

    refresh_token = _load_token_from_file(filepath)
    if refresh_token:
        try:
            # this will try to refresh, and raise if it fails
            return OpenIDRestClient(
                address=address,
                token_url=token_url,
                client_id=client_id,
                client_secret=client_secret,
                refresh_token=refresh_token,
                update_func=update_func,
                logger=logger,
                **kwargs,
            )
        except Exception:
            pass

    auth = OpenIDAuth(token_url)
    if not auth.provider_info:
        raise RuntimeError('Token service does not support .well-known discovery')
    if 'device_authorization_endpoint' not in auth.provider_info:
        raise RuntimeError('Device grant not supported by server')
    endpoint: str = auth.provider_info['device_authorization_endpoint']  # type: ignore

    device = CommonDeviceGrant()
    refresh_token = device.perform_device_grant(
        logger, endpoint, auth.token_url, client_id, client_secret, scopes
    )

    return OpenIDRestClient(
        address=address,
        token_url=token_url,
        client_id=client_id,
        client_secret=client_secret,
        refresh_token=refresh_token,
        update_func=update_func,
        logger=logger,
        **kwargs,
    )



import pytest
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key


@pytest.fixture(scope="session")
def gen_keys():
    priv = generate_private_key(65537, 2048)
    pub = priv.public_key()
    return (priv, pub)


@pytest.fixture(scope="session")
def gen_keys_bytes(gen_keys):
    priv, pub = gen_keys

    priv_pem = priv.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )
    pub_pem = pub.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    print(priv_pem, pub_pem)
    return (priv_pem, pub_pem)



"""Test REST Handler."""

# fmt:off
# pylint: skip-file

import json
import logging
from unittest.mock import MagicMock

import jwt.algorithms
import pytest
from rest_tools.server import (
    KeycloakUsernameMixin,
    OpenIDCookieHandlerMixin,
    OpenIDLoginHandler,
    RestHandler,
    RestHandlerSetup,
)
from rest_tools.utils.auth import Auth, OpenIDAuth
from tornado.web import Application, HTTPError

from .fixtures import gen_keys, gen_keys_bytes  # noqa: F401


def test_rest_handler_setup(requests_mock):
    ret = RestHandlerSetup({})
    assert ret['auth'] is None

    ret = RestHandlerSetup({'debug': True})
    assert ret['debug']

    ret = RestHandlerSetup({'auth': {'url': 'foo'}})
    assert ret['auth_url'] == 'foo'

    ret = RestHandlerSetup({'auth': {'secret': 'foo'}})
    assert isinstance(ret['auth'], Auth)

    # test audience and issuers
    ret = RestHandlerSetup({'auth': {'secret': 'foo', 'audience': 'bar', 'issuer': 'baz', 'issuers': ['baz']}})
    assert isinstance(ret['auth'], Auth)
    tok = ret['auth'].create_token('subj', expiration=20, payload={'aud': 'bar'})
    ret['auth'].validate(tok)

    requests_mock.get('http://foo/.well-known/openid-configuration', text=json.dumps({'token_endpoint': 'bar'}))
    ret = RestHandlerSetup({'auth': {'openid_url': 'http://foo'}})
    assert isinstance(ret['auth'], OpenIDAuth)
    assert requests_mock.called
    assert ret['auth_url'] == 'bar'

    ret = RestHandlerSetup({'rest_api': {'auth_key': 'foo'}})
    assert ret['module_auth_key'] == 'foo'


def test_rest_handler_initialize():
    rh = RestHandler()
    rh.initialize(debug=True)
    assert rh.debug


def test_rest_handler_get_current_user():
    a = Auth('secret')
    rh = RestHandler()
    rh.initialize(auth=a)
    rh.request = MagicMock()

    rh.request.headers = {}
    assert rh.get_current_user() is None

    token = a.create_token('subject', payload={'foo': 'bar'})
    rh.request.headers = {'Authorization': f'bearer {token}'}
    assert rh.get_current_user() == 'subject'
    assert rh.auth_data['foo'] == 'bar'
    assert rh.auth_key == token


def test_keycloak_username_mixin():
    auth_data = {}

    class Base:
        def get_current_user(self):
            self.auth_data = auth_data
            return 'foo'

    class A(KeycloakUsernameMixin, Base):
        pass

    test = A()

    assert test.get_current_user() is None

    auth_data['preferred_username'] = 'user'
    assert test.get_current_user() == 'user'

    del auth_data['preferred_username']
    auth_data['upn'] = 'user'
    assert test.get_current_user() == 'user'


def test_openid_cookie_handler_mixin():
    a = Auth('secret')

    class A(OpenIDCookieHandlerMixin, RestHandler):
        pass

    rh = A()
    rh.initialize(auth=a)
    rh.get_secure_cookie = MagicMock(return_value=b'')

    assert rh.get_current_user() is None

    token = a.create_token('subject', payload={'foo': 'bar'}).encode('utf-8')

    def get_token(name, *args, **kwargs):
        if name == 'access_token':
            return token
        elif name == 'refresh_token':
            return b''
        else:
            return None

    rh.get_secure_cookie = get_token

    assert rh.get_current_user() == 'subject'
    assert rh.auth_data['foo'] == 'bar'
    assert rh.auth_key == token


def test_openid_login_handler_initialize(requests_mock):
    handler = OpenIDLoginHandler()
    with pytest.raises(RuntimeError):
        handler.initialize('foo', 'bar')

    requests_mock.get('http://foo/.well-known/openid-configuration', text=json.dumps({
        'authorization_endpoint': 'http://foo/auth',
        'token_endpoint': 'http://foo/token',
        'end_session_endpoint': 'http://foo/logout',
        'userinfo_endpoint': 'http://foo/userinfo',
    }))
    ret = RestHandlerSetup({'auth': {'openid_url': 'http://foo'}})
    handler.initialize('foo', 'bar', **ret)

    assert handler._OAUTH_AUTHORIZE_URL == 'http://foo/auth'
    assert handler._OAUTH_ACCESS_TOKEN_URL == 'http://foo/token'
    assert handler._OAUTH_LOGOUT_URL == 'http://foo/logout'
    assert handler._OAUTH_USERINFO_URL == 'http://foo/userinfo'


@pytest.mark.asyncio
async def test_openid_login_handler_get_authenticated_user(gen_keys, gen_keys_bytes, requests_mock):  # noqa: F811
    handler = OpenIDLoginHandler()

    auth = Auth(gen_keys_bytes[0], pub_secret=gen_keys_bytes[1], algorithm='RS256')

    requests_mock.get('http://foo/.well-known/openid-configuration', text=json.dumps({
        'authorization_endpoint': 'http://foo/auth',
        'token_endpoint': 'http://foo/token',
        'end_session_endpoint': 'http://foo/logout',
        'userinfo_endpoint': 'http://foo/userinfo',
        'jwks_uri': 'http://foo/jwks',
    }))
    jwk = json.loads(jwt.algorithms.RSAAlgorithm.to_jwk(gen_keys[1]))
    jwk['kid'] = '123'
    logging.debug(jwk)
    requests_mock.get('http://foo/jwks', text=json.dumps({
        'keys': [jwk],
    }))
    ret = RestHandlerSetup({'auth': {'openid_url': 'http://foo'}})
    handler.initialize('foo', 'bar', **ret)

    access_token = auth.create_token('sub', headers={'kid': '123'})
    id_token = auth.create_token('sub', headers={'kid': '123'})

    user_info = {
        'id_token': id_token,
        'access_token': access_token,
        'expires_in': 3600,
    }

    async def fn(*args, **kwargs):
        ret = MagicMock()
        ret.body = json.dumps(user_info)
        return ret

    handler.get_auth_http_client = MagicMock()
    handler.get_auth_http_client.return_value.fetch = MagicMock(side_effect=fn)
    state = {}
    ret = await handler.get_authenticated_user('redirect', 'code', state)
    user_info_ret = user_info.copy()
    user_info_ret['id_token'] = auth.validate(id_token)
    assert ret == user_info_ret


def test_openid_login_handler_encode_decode_state(requests_mock):
    application = Application([], cookie_secret='secret')

    requests_mock.get('http://foo/.well-known/openid-configuration', text=json.dumps({
        'authorization_endpoint': 'http://foo/auth',
        'token_endpoint': 'http://foo/token',
        'end_session_endpoint': 'http://foo/logout',
        'userinfo_endpoint': 'http://foo/userinfo',
        'jwks_uri': 'http://foo/jwks',
    }))
    ret = RestHandlerSetup({'auth': {'openid_url': 'http://foo'}})
    ret['oauth_client_id'] = 'foo'
    ret['oauth_client_secret'] = 'bar'
    handler = OpenIDLoginHandler(application, MagicMock(), **ret)

    data = {'foo': 'bar'}
    state = handler._encode_state(data)
    data2 = handler._decode_state(state)
    assert data == data2


@pytest.mark.asyncio
async def test_openid_login_handler_get__no_body(gen_keys, gen_keys_bytes, requests_mock):  # noqa: F811
    application = Application([], cookie_secret='secret', login_url='/login', debug=True)

    # auth = Auth(gen_keys_bytes[0], pub_secret=gen_keys_bytes[1], algorithm='RS256')

    requests_mock.get('http://foo/.well-known/openid-configuration', text=json.dumps({
        'authorization_endpoint': 'http://foo/auth',
        'token_endpoint': 'http://foo/token',
        'end_session_endpoint': 'http://foo/logout',
        'userinfo_endpoint': 'http://foo/userinfo',
        'jwks_uri': 'http://foo/jwks',
    }))
    jwk = json.loads(jwt.algorithms.RSAAlgorithm.to_jwk(gen_keys[1]))
    jwk['kid'] = '123'
    logging.debug(jwk)
    requests_mock.get('http://foo/jwks', text=json.dumps({
        'keys': [jwk],
    }))
    ret = RestHandlerSetup({'auth': {'openid_url': 'http://foo'}})
    ret['oauth_client_id'] = 'foo'
    ret['oauth_client_secret'] = 'bar'

    # setup request
    request = MagicMock()
    handler = OpenIDLoginHandler(application, request, **ret)

    # first-time get
    handler.authorize_redirect = MagicMock()
    request.body = ''
    await handler.get()
    handler.authorize_redirect.assert_called()


@pytest.mark.asyncio
async def test_openid_login_handler_get__okay(gen_keys, gen_keys_bytes, requests_mock):  # noqa: F811
    application = Application([], cookie_secret='secret', login_url='/login', debug=True)

    auth = Auth(gen_keys_bytes[0], pub_secret=gen_keys_bytes[1], algorithm='RS256')

    requests_mock.get('http://foo/.well-known/openid-configuration', text=json.dumps({
        'authorization_endpoint': 'http://foo/auth',
        'token_endpoint': 'http://foo/token',
        'end_session_endpoint': 'http://foo/logout',
        'userinfo_endpoint': 'http://foo/userinfo',
        'jwks_uri': 'http://foo/jwks',
    }))
    jwk = json.loads(jwt.algorithms.RSAAlgorithm.to_jwk(gen_keys[1]))
    jwk['kid'] = '123'
    logging.debug(jwk)
    requests_mock.get('http://foo/jwks', text=json.dumps({
        'keys': [jwk],
    }))
    ret = RestHandlerSetup({'auth': {'openid_url': 'http://foo'}})
    ret['oauth_client_id'] = 'foo'
    ret['oauth_client_secret'] = 'bar'

    # setup request
    request = MagicMock()
    handler = OpenIDLoginHandler(application, request, **ret)

    # get with code
    token = auth.create_token('sub', headers={'kid': '123'})
    user_info = {
        'id_token': '{"id": "foo"}',
        'access_token': token,
        'expires_in': 3600,
    }

    async def fn2(*args, **Kwargs):
        return user_info

    handler.get_authenticated_user = MagicMock(side_effect=fn2)

    request.body = '{"code": "thecode", "state": "state"}'
    handler._decode_state = MagicMock(return_value={})
    handler.write = MagicMock()
    await handler.get()
    handler.write.assert_called()
    assert handler.write.call_args[0][0] == user_info


@pytest.mark.asyncio
async def test_openid_login_handler_get__with_error(gen_keys, gen_keys_bytes, requests_mock):  # noqa: F811
    application = Application([], cookie_secret='secret', login_url='/login', debug=True)

    # auth = Auth(gen_keys_bytes[0], pub_secret=gen_keys_bytes[1], algorithm='RS256')

    requests_mock.get('http://foo/.well-known/openid-configuration', text=json.dumps({
        'authorization_endpoint': 'http://foo/auth',
        'token_endpoint': 'http://foo/token',
        'end_session_endpoint': 'http://foo/logout',
        'userinfo_endpoint': 'http://foo/userinfo',
        'jwks_uri': 'http://foo/jwks',
    }))
    jwk = json.loads(jwt.algorithms.RSAAlgorithm.to_jwk(gen_keys[1]))
    jwk['kid'] = '123'
    logging.debug(jwk)
    requests_mock.get('http://foo/jwks', text=json.dumps({
        'keys': [jwk],
    }))
    ret = RestHandlerSetup({'auth': {'openid_url': 'http://foo'}})
    ret['oauth_client_id'] = 'foo'
    ret['oauth_client_secret'] = 'bar'

    # setup request
    request = MagicMock()
    handler = OpenIDLoginHandler(application, request, **ret)

    # get with error
    request.body = '{"error": true, "error_description": "the error"}'
    handler._decode_state = MagicMock(return_value={})
    handler.authorize_redirect = MagicMock()
    with pytest.raises(HTTPError, match='the error'):
        await handler.get()
    handler.authorize_redirect.assert_not_called()
